import * as Path from "path";
import {inject, injectable} from "inversify";
import {NxConfiguration} from "./NxConfiguration";
import {PersistenceService} from "../persistence/PersistenceService";
import {ConfigurationFileNotFoundError} from "../errors/ConfigurationFileNotFoundError";
import {LinkScope} from "./NxConfiguration";

const NxConfigurationFileName: string = "nx.json";

@injectable()
export class ConfigurationService {
    private _persistenceService: PersistenceService;

    constructor(@inject("PersistenceService") persistenceService: PersistenceService) {
        this._persistenceService = persistenceService;

    }

    public init(): Promise<NxConfiguration> {
        return new Promise((resolve) => {
            let configuration = new NxConfiguration();
            configuration.link = new LinkScope();
            configuration.bump = new LinkScope();
            resolve(configuration);
        });
    }

    public load(configurationFilePath: string): Promise<NxConfiguration> {
        return this._persistenceService.fileExists(configurationFilePath)
            .then((exists: boolean) => {
                if (exists === false) {
                    throw new ConfigurationFileNotFoundError(configurationFilePath);
                }
                return this._persistenceService.readFile(configurationFilePath)
                    .then((data: Buffer) => {
                        let configuration = <NxConfiguration> JSON.parse(data.toString("utf-8"));
                        if (configuration.link != undefined && configuration.link.files != undefined) {
                            configuration.link.files = this.resolveDirectories(configuration.link.files, Path.dirname(configurationFilePath));
                        }
                        if (configuration.bump != undefined && configuration.bump.files != undefined) {
                            configuration.bump.files = this.resolveDirectories(configuration.bump.files, Path.dirname(configurationFilePath));
                        }
                        return configuration;
                    });
            });
    }

    public search(directory: string): string {
        let directoryInfo = this._persistenceService.getDirectoryInfo(directory, false);
        if (directoryInfo.files != undefined) {
            for (let i = 0; i < directoryInfo.files.length; i++) {
                let file = directoryInfo.files[i];
                if (file.fileName === NxConfigurationFileName) {
                    return file.path;
                }
            }
        }
        return this.search(Path.dirname(directory));
    }

    public save(directory: string, configuration: NxConfiguration): Promise<boolean> {
        let fullPath = Path.join(directory, NxConfigurationFileName);
        return this._persistenceService.fileExists(fullPath)
            .then((exists: boolean) => {
                if (exists === true) {
                    return this._persistenceService.deleteFile(fullPath).then(() => {
                        return this._persistenceService.writeFile(fullPath, JSON.stringify(configuration, null, 4));
                    });
                }
                else {
                    return this._persistenceService.writeFile(fullPath, JSON.stringify(configuration, null, 4));
                }
            });
    }

    private resolveDirectories(directories: Array<string>, currentDirectory: string): Array<string> {
        return directories.map((dir) => {
            if (Path.isAbsolute(dir)) {
                return dir;
            }
            else {
                return Path.resolve(Path.join(currentDirectory, dir));
            }
        });
    }
}