import {injectable, inject} from "inversify";
import * as Path from "path";
import {PersistenceService} from "../persistence/PersistenceService";
import {LinkScope} from "../configuration/NxConfiguration";
import {PackageService} from "../node/PackageService";
import {LoggingServiceFactory} from "../logging/LoggingServiceFactory";
import {LoggingService} from "../logging/LoggingService";
import {PackageInfo} from "../node/PackageInfo";
import {NpmService} from "../node/NpmService";

@injectable()
export class LinkService {
    private _persistenceService: PersistenceService;
    private _packageService: PackageService;
    private _loggingService: LoggingService;
    private _npmService: NpmService;

    constructor(@inject("LoggingServiceFactory") loggingServiceFactory: LoggingServiceFactory,
                @inject("PersistenceService") persistenceService: PersistenceService,
                @inject("NpmService") npmService: NpmService,
                @inject("PackageService") packageService: PackageService) {
        this._loggingService = loggingServiceFactory.create("LinkService");
        this._persistenceService = persistenceService;
        this._npmService = npmService;
        this._packageService = packageService;
    }

    public link(scope: LinkScope): Promise<boolean> {
        this._loggingService.debug("start linking");
        let packageInfos = this._packageService.locatePackages(scope.files);
        this._loggingService.debug("packages located");
        packageInfos.forEach((packageInfo) => {
            this._loggingService.debug(`located ${packageInfo.name} (v. ${packageInfo.version}) at ${packageInfo.manifestPath}`, packageInfo);

            packageInfo.dependencies.forEach((dependency) => {
                this._loggingService.debug(`checking dependency ${dependency.name}`, dependency);
                let packageToLink = packageInfos.find((x) => x.name === dependency.name);
                if (packageToLink != undefined) {
                    this._loggingService.debug(`dependency ${dependency.name} matched to package ${packageToLink.name}`, {
                        dependency: dependency,
                        packageToLink: packageToLink
                    });
                    this._npmService.link(packageInfo, packageToLink);
                    this._loggingService.info(`linked ${packageToLink.name} from ${packageInfo.name}`);
                }
                else {
                    this._loggingService.debug(`no package matched for dependency ${dependency.name}`, dependency);
                }
            });

        });
        return Promise.resolve(true);
    }

    public unlink(scope: LinkScope): Promise<boolean> {
        this._loggingService.debug("start unlinking");
        let packageInfos = this._packageService.locatePackages(scope.files);
        this._loggingService.debug("packages located");
        packageInfos.forEach((packageInfo) => {
            this._loggingService.debug(`located ${packageInfo.name} (v. ${packageInfo.version}) at ${packageInfo.manifestPath}`, packageInfo);

            packageInfo.dependencies.forEach((dependency) => {
                this._loggingService.debug(`checking dependency ${dependency.name}`, dependency);
                let packageToUnlink = packageInfos.find((x) => x.name === dependency.name);
                if (packageToUnlink != undefined) {
                    this._loggingService.debug(`dependency ${dependency.name} matched to package ${packageToUnlink.name}`, {
                        dependency: dependency,
                        packageToLink: packageToUnlink
                    });
                    this.unlinkPackage(packageInfo, packageToUnlink);
                    this._loggingService.info(`unlinked ${packageToUnlink.name} from ${packageInfo.name}`);
                }
                else {
                    this._loggingService.debug(`no package matched for dependency ${dependency.name}`, dependency);
                }
            });

        });
        return Promise.resolve(true);
    }

    private unlinkPackage(packageInfo: PackageInfo, packageToUnlink: PackageInfo): void {

        let nodeModulesDirectoryPath = Path.join(Path.dirname(packageInfo.manifestPath), "node_modules");
        let nodeModulesDirectory = this._persistenceService.getDirectoryInfo(nodeModulesDirectoryPath, false);
        if (nodeModulesDirectory == undefined || nodeModulesDirectory.symbolicLinks == undefined) {
            this._loggingService.warn(`node modules folder not found ${nodeModulesDirectoryPath}`, nodeModulesDirectoryPath);
            return;
        }
        nodeModulesDirectory.symbolicLinks.forEach((moduleDirectory) => {

            let manifestPath = Path.join(moduleDirectory.path, "package.json");
            let manifest = this._packageService.loadManifest(manifestPath);

            if (manifest != undefined) {
                this._loggingService.warn(`manifest of ${manifest.name} loaded from ${manifestPath}`, manifest);
                if (packageToUnlink.name === manifest.name) {
                    this._persistenceService.unlinkDir(moduleDirectory.path);
                    this._loggingService.debug(`unlinked ${manifest.name} from ${packageInfo.name}`);
                }
            }
            else {
                this._loggingService.warn(`could not load manifest from ${manifestPath}`, manifestPath);
            }


        });
    }
}