import {injectable, inject} from "inversify";
import {PackageInfo} from "./PackageInfo";
import {FileInfo} from "../persistence/PersistenceService";
import {PersistenceService} from "../persistence/PersistenceService";
import {LoggingServiceFactory} from "../logging/LoggingServiceFactory";
import {LoggingService} from "../logging/LoggingService";

@injectable()
export class PackageService {

    private _loggingService: LoggingService;
    private _persistenceService: PersistenceService;

    constructor(@inject("LoggingServiceFactory") loggingServiceFactory: LoggingServiceFactory,
                @inject("PersistenceService") persistenceService: PersistenceService) {
        this._loggingService = loggingServiceFactory.create("PackageService");
        this._persistenceService = persistenceService;
    }

    public locatePackages(searchDirectories: Array<string>): Array<PackageInfo> {
        let packageFileInfos = this.getPackageFileInfos(searchDirectories);
        let packageInfos = packageFileInfos.map((packageFileInfo) => {
            let manifest = this.loadManifest(packageFileInfo.path);
            let packageInfo = new PackageInfo(manifest.name, packageFileInfo.path, manifest.version, manifest.dependencies, manifest.devDependencies);
            return packageInfo;
        });
        this._loggingService.debug(`created package infos (${packageInfos.length})`, packageInfos);
        return packageInfos;
    }

    public loadManifest(manifestPath: string): PackageManifest {
        this._loggingService.debug(`loading package manifest from ${manifestPath}`, manifestPath);
        let manifest: PackageManifest = null;
        try {
            manifest = <PackageManifest>require(manifestPath);
        }
        catch (error) {
            this._loggingService.error(`error while loading manifest from ${manifestPath}`, manifestPath);
            throw error;
        }
        this._loggingService.debug(`manifest loaded (${manifest.name})`, manifest);
        return manifest;
    }

    private getPackageFileInfos(searchDirectories: Array<string>): Array<FileInfo> {
        this._loggingService.debug(`searching for packages in ${searchDirectories.join()}`, searchDirectories);

        let directoryInfos = searchDirectories.map((x) => {
            return this._persistenceService.getDirectoryInfo(x, true, ["node_modules", ".git", ".idea"]);
        });
        this._loggingService.debug("retrieved directory infos", directoryInfos);

        let jsonFiles: Array<FileInfo> = new Array<FileInfo>();
        directoryInfos.forEach((x) => {
            let filteredFiles = this._persistenceService.getFileInfos(x, ".json");
            jsonFiles = jsonFiles.concat(filteredFiles);
        });

        this._loggingService.debug("filtered json files", jsonFiles);
        let packageFiles = jsonFiles.filter((x) => x.name === "package");

        this._loggingService.debug("filtered packages", packageFiles);
        return packageFiles;
    }

}

export interface PackageManifest {
    name: string;
    version: string;
    dependencies: { [key: string]: string };
    devDependencies: { [key: string]: string };
}