import * as Fs from "fs";
import * as Path from "path";
import * as Os from "os";
import {injectable} from "inversify";


@injectable()
export class PersistenceService {

    public ensureDirectory(directoryPath: string): Promise<boolean> {
        return new Promise((resolve, reject) => {
            Fs.exists(directoryPath, (exists: boolean) => {
                if (exists) {
                    resolve(true);
                }
                else {
                    Fs.mkdir(directoryPath, "0o777", (error) => {
                        if (error != undefined) {
                            reject(false);
                        }
                        else {
                            resolve(true);
                        }
                    });
                }
            });
        });
    }

    public getScopeDirectoryPath(): string {
        return Path.join(this.getNodevolutionDirectoryPath(), "scopes");
    }

    public getNodevolutionDirectoryPath(): string {
        let homeDirectoryPath = Os.homedir();
        return Path.join(homeDirectoryPath, ".nodevolution");
    }


    public getFileInfos(directoryInfo: DirectoryInfo): Array<FileInfo>;
    public getFileInfos(directoryInfo: DirectoryInfo, filter: string): Array<FileInfo>;
    public getFileInfos(directoryInfo: DirectoryInfo, filter?: string): Array<FileInfo> {

        if (directoryInfo.files == undefined && directoryInfo.directories == undefined) {
            return undefined;
        }

        let fileInfos: Array<FileInfo> = null;

        if (filter == undefined) {
            fileInfos = directoryInfo.files;
        }
        else if (directoryInfo.files != undefined) {
            fileInfos = directoryInfo.files.filter((x) => {
                return x.extension === filter;
            });
        }

        if (directoryInfo.directories != undefined) {
            directoryInfo.directories.forEach((subDirectory) => {
                let subFiles = this.getFileInfos(subDirectory, filter);
                if (subFiles != undefined) {
                    if (fileInfos == undefined) {
                        fileInfos = subFiles;
                    }
                    else {
                        fileInfos = fileInfos.concat(subFiles);
                    }

                }
            });
        }

        return fileInfos;
    }

    public getDirectoryInfo(directory: string, recursive: boolean): DirectoryInfo;
    public getDirectoryInfo(directory: string, recursive: boolean, exclude: Array<string>): DirectoryInfo;
    public getDirectoryInfo(directory: string, recursive: boolean, exclude?: Array<string>): DirectoryInfo {

        let rootFullPath = Path.resolve(directory);
        let entries = Fs.readdirSync(rootFullPath);

        let fileInfos: Array<FileInfo> = null;
        let directoryInfos: Array<DirectoryInfo> = null;
        let symbolicLinks: Array<SymbolicLink> = null;

        let relevantEntries = this.getRelevantWalkEntries(entries, exclude);

        relevantEntries.forEach((entry: string) => {
            let fullPath = Path.join(rootFullPath, entry)
            let stats = Fs.lstatSync(fullPath);

            if (stats.isFile()) {
                if (fileInfos == undefined) {
                    fileInfos = new Array<FileInfo>();
                }
                fileInfos.push(new FileInfo(fullPath));
            }
            else if (stats.isDirectory()) {
                let directoryInfo: DirectoryInfo = null;
                if (recursive) {
                    directoryInfo = this.getDirectoryInfo(fullPath, recursive, exclude);
                }
                else {
                    directoryInfo = new DirectoryInfo(fullPath, null, null, null);
                }
                if (directoryInfos == undefined) {
                    directoryInfos = new Array<DirectoryInfo>();
                }
                directoryInfos.push(directoryInfo);
            }
            else if (stats.isSymbolicLink()) {
                if(symbolicLinks == undefined){
                    symbolicLinks = new Array<SymbolicLink>();
                }
                symbolicLinks.push(new SymbolicLink(fullPath));
            }
        });
        return new DirectoryInfo(rootFullPath, directoryInfos, fileInfos, symbolicLinks);
    }

    private getRelevantWalkEntries(entries: Array<string>, exclude?: Array<string>): Array<string> {
        if (exclude == undefined) {
            return entries;
        }
        else {
            return entries.filter((x) => {
                let segments = Path.parse(x);
                return exclude.indexOf(segments.base) === -1 && exclude.indexOf(x) === -1;
            });
        }
    }

    public fileExists(path: string): Promise<boolean> {
        return new Promise((resolve, reject) => {
            Fs.access(path, (error) => {
                if (error != undefined) {
                    if (error.code === "EEXIST") {
                        resolve(true);
                    }
                    else {
                        resolve(false);
                    }
                }
                else {
                    resolve(true);
                }
            });

        });
    }

    public readFile(path: string): Promise<Buffer> {
        return new Promise((resolve, reject) => {
            Fs.readFile(path, (error: any, data: Buffer) => {
                if (error != undefined) {
                    reject(error);
                }
                else {
                    resolve(data);
                }
            });
        });
    }

    public deleteFile(path: string): Promise<boolean> {
        return new Promise((resolve, reject) => {
            Fs.unlink(path, (error) => {
                if (error != undefined) {
                    reject(error);
                }
                else {
                    resolve(true);
                }
            });
        });
    }

    public unlinkDir(path: string): Promise<boolean> {
        return new Promise((resolve, reject) => {
            Fs.unlink(path, (error) => {
                if (error != undefined) {
                    reject(error);
                }
                else {
                    resolve(true);
                }
            });
        });
    }

    public writeFile(path: string, content: string): Promise<boolean> {
        return new Promise((resolve, reject) => {
            Fs.writeFile(path, content, {encoding: "utf-8"}, (error) => {
                if (error != undefined) {
                    reject(error);
                }
                else {
                    resolve(true);
                }
            });
        });
    }

}


export class InfoBase {

    private _path: string;
    private _directory: string;
    private _name: string;

    public get path(): string {
        return this._path;
    }

    public get directory(): string {
        return this._directory;
    }

    public get name(): string {
        return this._name;
    }

    constructor(path: string) {
        this._path = path;
        let segments = Path.parse(path);
        this._directory = segments.dir;
        this._name = segments.name;
    }
}

export class SymbolicLink extends InfoBase{
}

export class DirectoryInfo extends InfoBase {
    private _files: Array<FileInfo>;
    private _directories: Array<DirectoryInfo>;
    private _symbolicLinks: Array<SymbolicLink>;

    public get files(): Array<FileInfo> {
        return this._files;
    }

    public get directories(): Array<DirectoryInfo> {
        return this._directories;
    }

    get symbolicLinks(): Array<SymbolicLink> {
        return this._symbolicLinks;
    }

    constructor(path: string, directories: Array<DirectoryInfo>, files: Array<FileInfo>, symbolicLinks: Array<SymbolicLink>) {
        super(path);
        this._directories = directories;
        this._files = files;
        this._symbolicLinks = symbolicLinks;
    }

}

export class FileInfo extends InfoBase {

    private _extension: string;
    private _fileName: string;

    public get extension(): string {
        return this._extension;
    }

    public get fileName(): string {
        return this._fileName;
    }

    constructor(path: string) {
        super(path);
        let segments = Path.parse(path);
        this._extension = segments.ext;
        this._fileName = `${segments.name}${segments.ext}`;
    }

}