import {inject, injectable} from "inversify";
import * as semver from "semver";
import {LoggingServiceFactory} from "../logging/LoggingServiceFactory";
import {LoggingService} from "../logging/LoggingService";
import {PackageInfo} from "../node/PackageInfo";
import {InvalidVersionError} from "../errors/InvalidVersionError";
import {Contract} from "./Contract";
import {VersionLevel} from "./VersionLevel";
import {PersistenceService} from "../persistence/PersistenceService";

@injectable()
export class VersionService {

    private _loggingService: LoggingService;
    private _bumpCommandFactory: BumpCommandFactory;
    private _persistenceService: PersistenceService;

    constructor(@inject("LoggingServiceFactory") loggingServiceFactory: LoggingServiceFactory,
                @inject("PersistenceService") persistenceService: PersistenceService) {
        this._persistenceService = persistenceService;
        this._loggingService = loggingServiceFactory.create("LinkService");
        this._bumpCommandFactory = new BumpCommandFactory();
    }

    public createVersinBumpInfos(initialPackage: PackageInfo, versionLevel: VersionLevel, packages: Array<PackageInfo>): Array<BumpInfo> {
        let tree = this.buildTreeItem(initialPackage, packages);
        let commands = Array<BumpCommandBase>();
        this.buildBumpCommands(tree, versionLevel, commands);
        return this.createBumpInfos(commands);
    }

    private createBumpInfos(commands: Array<BumpCommandBase>): Array<BumpInfo> {
        let bumpPackageCommands = commands.filter((x) => x instanceof BumpPackageCommand);
        let infos = bumpPackageCommands.map((x) => {
            let groupedByPackage = new Array<BumpCommandBase>();
            groupedByPackage = groupedByPackage.concat(commands.filter((y) => y.packageName === x.packageName));
            return new BumpInfo(x.packageInfo, groupedByPackage);
        });
        return infos.sort(this.sortByDependencies);
    }

    private sortByDependencies(that: BumpInfo, other: BumpInfo) {
        if (that.commands.length > other.commands.length) {
            return 1;
        }
        if (that.commands.length < other.commands.length) {
            return -1;
        }
        return 0;
    }

    public async persistChanges(bumpInfos: Array<BumpInfo>): Promise<void> {
        let persistencePromises: Array<Promise<void>> = bumpInfos.map((bumpInfo) => {
            if (bumpInfo.commands != undefined && bumpInfo.commands.length > 0) {
                return this.persistBumpInfoChanges(bumpInfo);
            }
            else {
                return Promise.resolve();
            }
        });
        return Promise.all(persistencePromises)
            .then(() => {
                return;
            });
    }

    private async persistBumpInfoChanges(bumpInfo: BumpInfo): Promise<void> {
        let content = (await this._persistenceService.readFile(bumpInfo.packageInfo.manifestPath)).toString();
        let packageFile = JSON.parse(content);
        bumpInfo.commands.forEach((cmd) => {
            if (cmd instanceof BumpPackageCommand) {
                packageFile.version = cmd.newVersion;
            }
            else if (cmd instanceof BumpDependencyCommand) {
                packageFile.dependencies[cmd.dependencyPackageName] = cmd.newVersion;
            }
            else if (cmd instanceof BumpDevDependencyCommand) {
                packageFile.devDependencies[cmd.dependencyPackageName] = cmd.newVersion;
            }
        });
        await this._persistenceService.writeFile(bumpInfo.packageInfo.manifestPath, JSON.stringify(packageFile, null, "\t"));
    }

    private buildBumpCommands(treeItem: DependencyTreeItem, versionLevel: VersionLevel, commands: Array<BumpCommandBase>) {
        let isNotInCommands = commands.find((x) => {
                return x instanceof BumpPackageCommand && x.packageName === treeItem.packageName;
            }) == undefined;

        if (isNotInCommands) {
            let bumpedVersion = this.getBumpedVersion(treeItem.packageInfo.version, versionLevel);
            commands.push(this._bumpCommandFactory.createBumpPackageCommand(treeItem.packageInfo, treeItem.packageInfo.version, bumpedVersion));
            if (treeItem.dependendItems != undefined) {
                treeItem.dependendItems.forEach((x) => {
                    this.buildBumpCommands(x, versionLevel, commands);
                    if (x.packageInfo.hasDependency(treeItem.packageName)) {
                        commands.push(this._bumpCommandFactory.createBumpDependencyCommand(x.packageInfo, treeItem.packageName, treeItem.packageInfo.version, bumpedVersion));
                    }
                    if (x.packageInfo.hasDevDependency(treeItem.packageName)) {
                        commands.push(this._bumpCommandFactory.createBumpDevDependencyCommand(x.packageInfo, treeItem.packageName, treeItem.packageInfo.version, bumpedVersion));
                    }
                });
            }
        }
    }

    private buildTreeItem(packageInfo: PackageInfo, packages: Array<PackageInfo>): DependencyTreeItem {
        let item = new DependencyTreeItem(packageInfo);
        packages.forEach((p) => {
            if (p.hasDependency(packageInfo.name)) {
                item.add(this.buildTreeItem(p, packages));
            }
        });
        return item;
    }

    private getBumpedVersion(version: string, level: VersionLevel): string {
        let validVersion = semver.valid(version);
        if (validVersion == undefined) {
            let error = new InvalidVersionError(version);
            this._loggingService.error(`${error.message} - ${version}`, error);
            throw error;
        }

        this._loggingService.debug(`passed version is a valid semver version - ${validVersion}`, validVersion);
        let semverVersionLevel = this.getSemverVersionLevel(level);
        let incrementedVersion = semver.inc(validVersion, semverVersionLevel);
        this._loggingService.debug(`incremented version ${validVersion} => ${incrementedVersion}`);

        return incrementedVersion;
    }

    public getVersionLevel(versionLevel: string): VersionLevel {
        let loweredVersionLevel = versionLevel.toLowerCase();
        switch (loweredVersionLevel) {
            case "major":
                return VersionLevel.Major;
            case "minor":
                return VersionLevel.Minor;
            case "patch":
                return VersionLevel.Patch;
            case "premajor":
                return VersionLevel.Premajor;
            case "preminor":
                return VersionLevel.Preminor;
            case "prepatch":
                return VersionLevel.Prepatch;
            case "prerelease":
                return VersionLevel.Prerelease;
            default:
                return undefined;
        }
    }

    private getSemverVersionLevel(level: VersionLevel): Contract.Semver.VersionLevel {
        switch (level) {
            case VersionLevel.Major:
                return "major";
            case VersionLevel.Minor:
                return "minor";
            case VersionLevel.Patch:
                return "patch";
            case VersionLevel.Premajor:
                return "premajor";
            case VersionLevel.Preminor:
                return "preminor";
            case VersionLevel.Prepatch:
                return "prepatch";
            case VersionLevel.Prerelease:
                return "prerelease";
        }
    }


}

export class BumpInfo {


    get commands(): Array<BumpCommandBase> {
        return this._commands;
    }

    get packageName(): string {
        return this._packageInfo.name;
    }

    get packageInfo() {
        return this._packageInfo;
    }

    private _commands: Array<BumpCommandBase>;
    private _packageInfo: PackageInfo;

    constructor(packageInfo: PackageInfo, commands: Array<BumpCommandBase>) {
        this._packageInfo = packageInfo;
        this._commands = commands;
    }

}

export class BumpCommandFactory {

    public createBumpPackageCommand(packageInfo: PackageInfo, version: string, newVersion: string): BumpPackageCommand {
        return new BumpPackageCommand(packageInfo, version, newVersion);
    }

    public createBumpDependencyCommand(packageInfo: PackageInfo, dependencyPackageName: string, version: string, newVersion: string): BumpDependencyCommand {
        return new BumpDependencyCommand(packageInfo, dependencyPackageName, version, newVersion);
    }

    public createBumpDevDependencyCommand(packageInfo: PackageInfo, dependencyPackageName: string, version: string, newVersion: string): BumpDevDependencyCommand {
        return new BumpDevDependencyCommand(packageInfo, dependencyPackageName, version, newVersion);
    }

}

export class DependencyTreeItem {


    get dependendItems(): Array<DependencyTreeItem> {
        return this._dependendItems;
    }

    get packageName(): string {
        return this._packageInfo.name;
    }

    get packageInfo(): PackageInfo {
        return this._packageInfo;
    }

    private _dependendItems: Array<DependencyTreeItem>;
    private _packageInfo: PackageInfo;

    constructor(packageInfo: PackageInfo) {
        this._packageInfo = packageInfo;
    }

    public add(item: DependencyTreeItem): void {
        if (this._dependendItems == undefined) {
            this._dependendItems = new Array<DependencyTreeItem>();
        }
        this._dependendItems.push(item);
    }
}

export abstract class BumpCommandBase {


    private _version: string;
    private _newVersion: string;
    private _packageInfo: PackageInfo;

    get packageName(): string {
        return this._packageInfo.name;
    }

    get version(): string {
        return this._version;
    }

    get newVersion(): string {
        return this._newVersion;
    }

    get packageInfo() {
        return this._packageInfo;
    }

    constructor(packageInfo: PackageInfo, version: string, newVersion: string) {
        this._packageInfo = packageInfo;
        this._version = version;
        this._newVersion = newVersion;
    }

}

export class BumpPackageCommand extends BumpCommandBase {
}

export class BumpDependencyCommand extends BumpCommandBase {

    private _dependencyPackageName: string;

    get dependencyPackageName(): string {
        return this._dependencyPackageName;
    }

    constructor(packageInfo: PackageInfo, dependencyPackageName: string, version: string, newVersion: string) {
        super(packageInfo, version, newVersion);
        this._dependencyPackageName = dependencyPackageName;
    }
}

export class BumpDevDependencyCommand extends BumpCommandBase {

    private _dependencyPackageName: string;

    get dependencyPackageName(): string {
        return this._dependencyPackageName;
    }

    constructor(packageInfo: PackageInfo, dependencyPackageName: string, version: string, newVersion: string) {
        super(packageInfo, version, newVersion);
        this._dependencyPackageName = dependencyPackageName;
    }
}
